from pathlib import Path
import typer
from rich.console import Console
import pandas as pd

from .apparatus import Doc
from .export import export_variants_to_excel, import_classifications_from_dataframe
from .classification import classify as classify_fn
from .evaluation import evaluate_docs
from .classification import DEFAULT_MODEL_ID
from .validation import validate as validate_fn
from .prompts import build_preamble

console = Console()
error_console = Console(stderr=True, style="bold red")


app = typer.Typer(pretty_exceptions_enable=False)
    


def get_output_path(doc:Path, output:Path, inplace:bool) -> Path:
    """ Checks if the output path should be replaced with the input doc. """
    if output and inplace:
       raise typer.BadParameter("You cannot use both an output path and --inplace/-i at the same time.")
    if not output and not inplace:
        raise typer.BadParameter("You must provide either an output path or use --inplace/-i.")
    
    if inplace:
        output = doc
    
    return output


@app.command()
def classify(
    doc:Path=typer.Argument(..., help="The path to the TEI XML document to classify."),
    output:Path=typer.Argument(None, help="The path to the output TEI XML file."),
    inplace: bool = typer.Option(False, "--inplace", "-i", help="Overwrite the input file."),
    verbose:bool=typer.Option(False, help="Print verbose output."),
    api_key:str=typer.Option("", help="API key for the LLM."),
    llm:str=typer.Option(DEFAULT_MODEL_ID, help="ID of the language model to use."),
    temperature:float=typer.Option(0.1, help="Temperature for sampling from the language model."),
    prompt_only:bool=typer.Option(False, help="Only print the prompt and not classify."),
    examples:int=typer.Option(10, help="Number of examples to include in the prompt."),
):
    """
    Classifies relations in TEI documents.
    """
    doc = Doc(doc)
    output = get_output_path(doc, output, inplace)

    return classify_fn(
        doc=doc, 
        output=output, 
        verbose=verbose, 
        api_key=api_key, 
        llm=llm, 
        temperature=temperature,
        prompt_only=prompt_only, 
        examples=examples, 
        console=console,
    )


@app.command()
def classified_pairs(
    doc:Path=typer.Argument(..., help="The path to the TEI XML document with the classifications."),
):
    """ Print classified pairs in a document. """
    doc = Doc(doc)
    doc.print_classified_pairs(console)


@app.command()
def html(
    doc:Path=typer.Argument(..., help="The path to the TEI XML document to render as HTML."),
    output:Path=typer.Argument(..., help="The path to the output HTML file."),
    all_apps:bool=typer.Option(False, help="Whether or not to use all variation unit `app` elements. By default it shows only non-redundant pairs of readings."),
):    
    """ Renders the variation units of a TEI document as HTML. """
    doc = Doc(doc)
    doc.render_html(output, all_apps=all_apps)


@app.command()
def gui(
    doc:Path=typer.Argument(..., help="The path to the TEI XML document to classify."),
    output:Path=typer.Argument(None, help="The path to the output TEI XML file."),
    inplace: bool = typer.Option(False, "--inplace", "-i", help="Overwrite the input file."),
    debug:bool=True,
    use_reloader:bool=False,
    all_apps:bool=typer.Option(False, help="Whether or not to use all variation unit `app` elements. By default it shows only non-redundant pairs of readings."),
):
    """ Starts a Flask app to view and classify a TEI document. """
    doc = Doc(doc)
    output = get_output_path(doc, output, inplace)
    flask_app = doc.flask_app(output, all_apps=all_apps)
    flask_app.run(debug=debug, use_reloader=use_reloader)


@app.command()
def evaluate(
    predicted:Path=typer.Argument(..., help="The path to the TEI XML document with predictions from Rdgai to evaluate."),
    ground_truth:Path=typer.Argument(..., help="The path to the input TEI XML document to use as the ground truth for evaluation."),
    confusion_matrix:Path=typer.Option(None, help="Path to write the confusion matrix plot as a CSV file."),
    confusion_matrix_plot:Path=typer.Option(None, help="Path to write the confusion matrix plot as an HTML file."),
    report:Path=typer.Option(None, help="Path to write the report."),
):
    """ Evaluates the classifications in a predicted document against a ground truth document. """
    predicted = Doc(predicted)
    ground_truth = Doc(ground_truth)
    
    evaluate_docs(predicted, ground_truth, confusion_matrix=confusion_matrix, confusion_matrix_plot=confusion_matrix_plot, report=report)


@app.command()
def validate(
    ground_truth:Path=typer.Argument(..., help="The path to the input TEI XML document to use as the ground truth for evaluation."),
    output:Path=typer.Argument(..., help="The path to the output TEI XML file."),
    proportion:float=typer.Option(0.5, help="Proportion of classified pairs to use for validation."),
    api_key:str=typer.Option("", help="API key for the LLM."),
    llm:str=typer.Option(DEFAULT_MODEL_ID, help="ID of the language model to use."),
    temperature:float=typer.Option(0.1, help="Temperature for sampling from the language model."),
    examples:int=typer.Option(10, help="Number of examples to include in the prompt."),
    confusion_matrix:Path=typer.Option(None, help="Path to write the confusion matrix plot as a CSV file."),
    confusion_matrix_plot:Path=typer.Option(None, help="Path to write the confusion matrix plot as an HTML file."),
    seed:int=typer.Option(42, help="Seed for random sampling of validation pairs."),
    report:Path=typer.Option(None, help="Path to write the report."),
):
    """ Takes a ground truth document, chooses a proportion of classified pairs to validate against and outputs a report. """
    ground_truth = Doc(ground_truth)
    
    validate_fn(
        ground_truth, 
        output,
        llm=llm,
        api_key=api_key,
        examples=examples,
        seed=seed,
        temperature=temperature,
        proportion=proportion,
        confusion_matrix=confusion_matrix, 
        confusion_matrix_plot=confusion_matrix_plot, 
        report=report,
    )


@app.command()
def clean(
    doc:Path=typer.Argument(..., help="The path to the TEI XML document to clean."),
    output:Path=typer.Argument(None, help="The path to the output TEI XML file."),
    inplace: bool = typer.Option(False, "--inplace", "-i", help="Overwrite the input file."),
):
    """ Cleans a TEI XML file for common errors. """
    doc = Doc(doc)
    output = get_output_path(doc, output, inplace)
    doc.clean(output=output)


@app.command()
def export(
    doc:Path=typer.Argument(..., help="The path to the TEI XML document to export."),
    output:Path=typer.Argument(..., help="The path to the output Excel file."),
):
    """ Exports pairs of readings with classifications from a TEI document to an Excel spreadsheet. """
    doc = Doc(doc)
    export_variants_to_excel(doc, output)


@app.command()
def import_classifications(
    doc:Path=typer.Argument(..., help="The path to the base TEI XML document to use for importing the classifications from Excel."),
    spreadsheet:Path=typer.Argument(..., help="The path to the Excel file to import."),
    output:Path=typer.Argument(None, help="The path to the output TEI XML file."),
    inplace: bool = typer.Option(False, "--inplace", "-i", help="Overwrite the input file."),
    responsible:str=typer.Option("", help="The responsible party for the classifications. By default it is the name of the spreadsheet."),
):
    """ Imports classifications from a spreadsheet into a TEI document. """
    doc = Doc(doc)
    output = get_output_path(doc, output, inplace)

    if spreadsheet.suffix == ".xlsx":
        variants_df = pd.read_excel(spreadsheet, sheet_name="Variants", keep_default_na=False)
    elif spreadsheet.suffix == ".csv":
        variants_df = pd.read_csv(spreadsheet, keep_default_na=False)

    # TODO add responsible to TEI header
    responsible = responsible or spreadsheet.stem
    responsible = responsible.replace(" ", "_")
    if not responsible.startswith("#"):
        responsible = "#" + responsible
    
    import_classifications_from_dataframe(doc, variants_df, output, responsible=responsible)


@app.command()
def prompt_preamble(
    doc:Path=typer.Argument(..., help="The path to the TEI XML document to classify."),
    examples:int=typer.Option(10, help="Number of examples to include in the prompt."),
):
    """ Prints the prompt preamble for a TEI document for a given number of examples. """
    doc = Doc(doc)
    template = build_preamble(doc, examples)
    print(template)



import random
from pathlib import Path
from rich.console import Console
import llmloader

from .apparatus import Doc, Pair
from .classification import classify, DEFAULT_MODEL_ID
from .evaluation import evaluate_docs


def validate(
    ground_truth:Doc,
    output:Path,
    validation_pairs:list[Pair]|None=None,
    verbose:bool=False,
    proportion:float=0.5,
    seed:int=42,
    api_key:str="",
    llm:str=DEFAULT_MODEL_ID,
    temperature:float=0.1,
    examples:int=10,
    console:Console|None=None,
    confusion_matrix:Path|None=None,
    confusion_matrix_plot:Path|None=None,
    report:Path|None=None,
):
    """
    Partitions the classified pairs in the document and uses a proportion for examples and the remainder for classification.
    Then it evaluates the classifications and writes a report.
    """
    ground_truth.write(output)
    doc = Doc(output)

    llm = llmloader.load(model=llm, api_key=api_key, temperature=temperature)

    # Find pairs to classify
    if not validation_pairs:
        random.seed(seed)
        classified_pairs = doc.get_classified_pairs(redundant=False)
        validation_pairs = random.sample(classified_pairs, int(len(classified_pairs) * proportion))
        validation_pairs = sorted(validation_pairs, key=lambda pair: (str(pair.active.app), pair.active.n, pair.passive.n))

    # Remove classifications from validation pairs
    for pair in validation_pairs:
        pair.remove_all_types()

    # Classify pairs
    classify(
        doc, 
        output,
        pairs=validation_pairs,
        verbose=verbose,
        llm=llm,
        examples=examples,
        console=console,
    )

    # Evaluate classifications
    evaluate_docs(
        doc,
        ground_truth,
        pairs=validation_pairs,
        confusion_matrix=confusion_matrix,  
        confusion_matrix_plot=confusion_matrix_plot,
        report=report,
        examples=examples,
        llm=llm,
    )




from dataclasses import dataclass
from langchain_core.runnables import Runnable

@dataclass
class CategoryParser(Runnable):
    relation_type_names: list[str]

    def invoke(self, llm_output:str, *args, **kwargs) -> tuple[str, str]:
        """
        Parses the output of a language model to category and justification.
        """
        llm_output = llm_output.strip()
        if "-----" in llm_output:
            llm_output = llm_output[:llm_output.find("-----")]

        category = ""
        justification = ""
        minimim_index = len(llm_output)
        for relation_type in self.relation_type_names:
            if relation_type in llm_output:
                index = llm_output.find(relation_type)
                if index < minimim_index:
                    minimim_index = index
                    category = relation_type
                    justification = llm_output[index + 1:]
                    justification_index = justification.find("\n")
                    justification = justification[justification_index + 1:].strip()
                    
        return category, justification




from langchain_core.messages import HumanMessage, AIMessage, SystemMessage
from langchain_core.prompts.chat import ChatPromptTemplate
from .apparatus import Pair, Doc


def select_spaced_elements(lst:list, k:int) -> list:
    n = len(lst)
    if k >= n:
        return lst
    if k == 1:
        return [lst[0]]
    if k == 2:
        return [lst[0], lst[-1]]
    
    # Always include the first and last element
    indices = [0] + [int((n - 1) * i / (k - 1)) for i in range(1, k - 1)] + [n - 1]
    
    return [lst[i] for i in indices]


# def build_template(app:App, examples:int=10) -> ChatPromptTemplate:
#     doc = app.doc
    
#     readings = app.readings

#     system_message = f"You are an academic who is an expert in textual criticism in {language}."

#     human_message = (
#         f"I am analyzing textual variants in a {language} document.\n"
#         "I want to classify the types of changes made to readings as the text was copied.\n"
#         "If reading 1 were to change to reading 2, what type of change would that be?\n"
#     )
    
#     # Add the categories to the message
#     relation_categories = doc.relation_types.values()
#     human_message += f"\nHere are {len(relation_categories)} possible categories for the types of changes in the text:\n"     
#     for category in relation_categories:
#         human_message += f"{category.str_with_description()}\n"
#         instance_strings = sorted(set(instance.reading_transition_str() for instance in category.pairs if not instance.rdgai_responsible()), key=len)
#         for instance_string in select_spaced_elements(instance_strings, examples):
#             human_message += f"\te.g. {instance_string}\n"

#     # Add the context to the message
#     human_message += f"\nI am variation unit is marked as {app.text_with_signs()} in this text:\n"
#     human_message += f"{app.text_in_context()}\n"

#     # Add the readings to the message        
#     human_message += f"\nHere are the {len(readings)} readings at that variation unit. Remember them so you can analyze them later:\n"
#     for reading in readings:
#         reading_text = reading.text or "OMITTED"
#         human_message += f"{reading.n}: {reading_text}\n"

#     # Describe output format
#     human_message += (
#         f"\nYou need to output the classifications of the readings in the following format.\n"
#         # "If going reading 1 to reading 2 was the same category of change, you would output the following:\n"
#         # "reading_2_id <-> reading_1_id = category : justification\n"

#         "If reading 1 were to change to reading 2, you would output the following:\n"
#         "reading_1_id → reading_2_id = category : justification\n"
#         "You also should output classificaitons from reading 2 to reading 1. It could be the same category or a different one:\n"
#         "reading_2_id → reading_1_id = category : justification\n"
        
#         "For the justification, give a one sentence explanation of why the change is classified as that category.\n"
#         "Separate the result for each pair of readings with a newline and do not add extra new line characters.\n"
#     )

#     # Add the classified relation examples to the message
#     human_message += f"\nGiven the readings above, you can output the following pairs of relations:\n"
#     for i, reading1 in enumerate(readings):
#         for j in range(i + 1, len(readings)):
#             reading2 = readings[j]
#             human_message += f"{reading1.n} → {reading2.n}\n"

#     human_message += "\nIf the change between one of those pairs of readings does not fit one of the categories, then do not output anything for that pair.\n"
#     human_message += f"\nWhen you are finished, output 5 hyphens: '-----'.\n"

#     ai_message = "Certainly, classifications for combinations of the readings are:"

#     template = ChatPromptTemplate.from_messages(messages=[
#         SystemMessage(system_message),
#         HumanMessage(human_message),
#         AIMessage(ai_message),        
#     ])
#     return template


def build_system_message(doc:Doc) -> str:
    return f"You are an academic who is an expert in textual criticism in {doc.language}."


def build_preamble(doc:Doc, examples:int=10) -> str:
    relation_categories = doc.relation_types.values()
    human_message = (
        f"I am analyzing textual variants in a document written in {doc.language}.\n"
        f"There are {len(relation_categories)} possible categories for the types of changes in the text that a scribe could make to change one variant reading to another.\n"
        "Here are the categories with their descriptions. Remember them because you will use this information later:\n"
    )
    for category in relation_categories:
        human_message += f"{category.str_with_description()}\n"

    human_message += "\n"
    human_message += f"Here are examples of these categories The word 'OMIT' indicates the absence of text in this reading. If there is a justification describing the choice of category then it is added in square brackets []:\n"
    for category in relation_categories:
        representative_category_exampes = category.representative_examples(examples)
        if len(representative_category_exampes) == 0:
            continue
        human_message += f"{category}:\n"        
        for pair in representative_category_exampes:
            human_message += f"\te.g. {pair.reading_transition_str()}"
            if pair.has_description():
                human_message += f" [{pair.get_description()}]"
            human_message += "\n"

    human_message += "\n"
    human_message += "I will give you a two variant readings. On the first line of your response, provide the correct category name for changing from the first reading to the second reading.\n"
    human_message += "Do not provide any other text than the name of the category on the first line.\n"
    human_message += "Then on a second line, give a justification of your decision according the definitions of the categories provided and similar examples.\n"
    human_message += f"When you are finished, output 5 hyphens: '-----'.\n"

    return human_message


def build_template(pair:Pair, examples:int=10) -> ChatPromptTemplate:
    app = pair.app
    doc = app.doc

    system_message = build_system_message(doc)
    human_message = build_preamble(doc, examples)    

    human_message += f"\nThe variation unit you need to classify is marked as {app.text_with_signs(pair.active.text)} in this text:\n"
    human_message += f"{app.text_in_context(pair.active.text)}\n"

    active_reading_text = app.text_with_signs(str(pair.active))
    passive_reading_text = app.text_with_signs(str(pair.passive))
    human_message += f"\nWhat category would best describe a change from {active_reading_text} to {passive_reading_text}?\n"

    relation_categories = doc.relation_types.values()
    relation_categories_list = ", ".join(str(category) for category in relation_categories)
    human_message += f"Respond with one of these categories: {relation_categories_list}\n"
    human_message += f"On the second line, provide a justification for your decision."

    ai_message = f"Certainly, the category for changing from {active_reading_text} to {passive_reading_text} is:"

    template = ChatPromptTemplate.from_messages(messages=[
        SystemMessage(system_message),
        HumanMessage(human_message),
        AIMessage(ai_message),        
    ])
    return template


def build_review_prompt(
    doc:Doc,
    correct_items:list["EvalItem"],
    incorrect_items:list["EvalItem"],
    examples:int=10,
):
    prompt_preamble = build_preamble(doc, examples=examples)

    system_message = f"You are an expert prompt engineer with expertise in {doc.language}.\n"

    human_message = "Please review the prompt based on results compared with the ground truth.\n"
    human_message += f"Here is the prompt:\n```\n{prompt_preamble}```\n"
    human_message += "----\nHere are the correctly classified items:\n"
    for item in correct_items:
        human_message += f" - { item.app_id }: {item.reading_transition_str} in `{item.text_in_context}` was correctly classified as {', '.join(item.predicted)} with this as the justification: {item.description}"
        if item.ground_truth_description:
            human_message += f" The justification for the ground truth {', '.join(item.ground_truth)} is: {item.ground_truth_description}"
        human_message += "\n"

    human_message += "----\nHere are the incorrectly classified items:\n"
    for item in incorrect_items:
        human_message += f" - { item.app_id }: {item.reading_transition_str} in `{item.text_in_context}` was incorrectly classified as {', '.join(item.predicted)} when it should have been {', '.join(item.ground_truth)}. This was the justification given: {item.description}"
        if item.ground_truth_description:
            human_message += f" The justification for the ground truth {', '.join(item.ground_truth)} is: {item.ground_truth_description}"
        human_message += "\n"

    human_message += "----\nPlease provide feedback on the prompt. Are the text be editing to improve the accuracy of the results? Are there examples in the test set where the ground truth label is incorrect? If so, then list all the problematic cases with their locations.\n"

    template = ChatPromptTemplate.from_messages(messages=[
        SystemMessage(system_message),
        HumanMessage(human_message),
    ])

    return template


from functools import cache
from pathlib import Path

@cache
def read_language_codes_yaml_cached() -> dict[str, str]:
    path = Path(__file__).parent / "data/language-subtag-registry.yaml"

    assert path.exists(), f"File not found: {path}"

    import yaml    
    with open(path, encoding='utf8') as f:
        result = yaml.safe_load(f)
    return result


def convert_language_code(code:str) -> str:
    codes = read_language_codes_yaml_cached()
    if code in codes:
        return codes[code]
    
    return code




class Mapper():
    def __init__(self):
        self.key_to_object = {}
        self.object_to_key = {}
    
    def key(self, object):
        if object in self.object_to_key:
            return self.object_to_key[object]
        
        key = str(object)
        # check if key is already in use and if it is, then make unique by adding number
        if key in self.key_to_object:
            i = 2
            while f"{key}_{i}" in self.key_to_object:
                i += 1
            key = f"{key}_{i}"

        self.key_to_object[key] = object
        self.object_to_key[object] = key
        return key

    def obj(self, key):
        return self.key_to_object.get(key)


from pathlib import Path
from openpyxl import Workbook
from openpyxl.worksheet.datavalidation import DataValidation
from openpyxl.styles import Font
import pandas as pd

from .apparatus import Doc


def export_variants_to_excel(doc:Doc, output:Path):
    """ Export the variants to an Excel file."""
    wb = Workbook()
    header_font = Font(bold=True)

    relation_types = doc.relation_types

    # Rename the default sheet
    ws = wb.active
    ws.title = 'Variants'

    headers = [
        'App ID', 'Context', 
        'Active Reading ID', 'Passive Reading ID',
        'Active Reading Text', 'Passive Reading Text', 
        'Description', 'Relation Type(s)',
    ]

    for col_num, header in enumerate(headers, start=1):  # Start from column A (1)
        cell = ws.cell(row=1, column=col_num)
        cell.value = header
        cell.font = header_font

    current_row = 2
    for app in doc.apps:
        for pair in app.non_redundant_pairs:
            ws[f'A{current_row}'] = str(app)
            ws[f'B{current_row}'] = app.text_in_context()
            ws[f'C{current_row}'] = pair.active.n
            ws[f'D{current_row}'] = pair.passive.n
            ws[f'E{current_row}'] = pair.active.text
            ws[f'F{current_row}'] = pair.passive.text
            ws[f'G{current_row}'] = pair.get_description()
            
            for relation_type_index, relation_type in enumerate(pair.types):
                column = ord('H') + relation_type_index
                ws[f'{chr(column)}{current_row}'] = str(relation_type)
            
            current_row += 1

    data_val = DataValidation(type="list",formula1=f'"{",".join(relation_types.keys())}"')
    ws.add_data_validation(data_val)

    max_relation_types = max(10, max(len(pair.types) for app in doc.apps for pair in app.pairs))
    end_column = chr(ord('H') + max_relation_types - 1)

    data_val.add(f"H2:{end_column}{current_row}")

    # Create new sheet with descriptions of categories and counts
    categories_worksheet = wb.create_sheet('Categories')

    # Add a header to the "Category" column
    headers = ['Category', 'Inverse', 'Count', 'Inverse Count', 'Total', 'Description']
    for col_num, header in enumerate(headers, start=1): 
        cell = categories_worksheet.cell(row=1, column=col_num)
        cell.value = header
        cell.font = header_font

    # Populate the categories from relation_types.keys()
    for idx, category in enumerate(relation_types.values(), start=2):  # Start from row 2
        category_name = str(category)
        inverse_name = str(category.inverse) if category.inverse else category_name
        categories_worksheet[f'A{idx}'] = category_name
        categories_worksheet[f'B{idx}'] = inverse_name
        categories_worksheet[f'C{idx}'] = f'=COUNTIF(Variants!G:{end_column}, "{category_name}")'
        categories_worksheet[f'D{idx}'] = f'=COUNTIF(Variants!G:{end_column}, "{inverse_name}")'
        categories_worksheet[f'E{idx}'] = f'=SUM(C{idx}:D{idx})'
        categories_worksheet[f'F{idx}'] = category.description

    wb.save(output)


def import_classifications_from_dataframe(doc:Doc, variants_df:pd.DataFrame, output:Path, responsible:str|None=None):
    variants_df.fillna('', inplace=True)
    apps_dict = {str(app): app for app in doc.apps}
    relation_types = doc.relation_types
    for _, row in variants_df.iterrows():
        app_id = row['App ID']
        active_reading_id = row['Active Reading ID']
        passive_reading_id = row['Passive Reading ID']
        description = row['Description'].strip() if 'Description' in row else None
        app = apps_dict[app_id]

        types = set(row[key] for key in row.keys() if (key.startswith('Relation Type') or key.startswith('Unnamed: ')) and row[key])
        # for type in types:
        #     assert type in relation_types, f'{type} not in {relation_types.keys()}'
        types = set(relation_types[type] for type in types if type in relation_types)

        for pair in app.pairs:
            if str(pair.active.n) == str(active_reading_id) and str(pair.passive.n) == str(passive_reading_id):
                # Add relations
                for type in types - pair.types:
                    pair.add_type_with_inverse(type, responsible=responsible)
            
                # Remove relations
                for type in pair.types - types:
                    pair.remove_type_with_inverse(type)

                if description:
                    pair.add_description(description)
                elif description == "":
                    # remove description if it is an empty string
                    # don't do anything if description is 'None'
                    pair.remove_description()

    doc.write(output)        



from typing import Optional
from pathlib import Path
from dataclasses import dataclass, field
from lxml.etree import _Element as Element
from lxml.etree import _ElementTree as ElementTree
from lxml import etree as ET
from rich.console import Console
import functools
import Levenshtein
import numpy as np

# from .relations import Relation, get_reading_identifier
from .tei import read_tei, find_elements, extract_text, find_parent, find_element, write_tei, make_nc_name, get_language, get_reading_identifier
from .mapper import Mapper

@dataclass
class Reading():
    element: Element
    app:"App"
    n: str  = field(default=None)
    text: str  = field(default=None)
    witnesses: list[str] = field(default_factory=list)

    def __post_init__(self):
        self.n = get_reading_identifier(self.element)
        self.text = extract_text(self.element).strip()
        self.witnesses = self.element.attrib.get("wit", "").split()

    def __str__(self):
        return self.text or 'OMIT'
    
    def witnesses_str(self) -> str:
        return " ".join(self.witnesses)

    def __hash__(self):
        return hash(self.element)


@dataclass
class RelationType():
    element: Element
    name: str
    description: str
    inverse: Optional['RelationType'] = None
    pairs: set['Pair'] = field(default_factory=set)

    def __str__(self):
        return self.name
    
    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other):
        if isinstance(other, RelationType):
            return (self.name, self.element, self.description) == (other.name, other.element, other.description)
        return False

    def __hash__(self):
        return hash((self.name, self.element, self.description))
    
    def str_with_description(self) -> str:
        result = self.name
        if self.description:
            result += f": {self.description}"
        return result

    def pairs_sorted(self, exclude_rdgai:bool = False) -> list['Pair']:
        pairs = sorted(self.pairs, key=lambda pair: (str(pair.active.app), pair.active.n, pair.passive.n))
        if exclude_rdgai:
            pairs = [pair for pair in pairs if not pair.rdgai_responsible()]
        return pairs

    def get_inverse(self) -> 'RelationType':
        return self.inverse if self.inverse else self
    
    @functools.lru_cache(maxsize=None)
    def representative_examples(self, k:int, random_state:int=42) -> list['Pair']:
        
        def find_representative_examples(pairs_list:list[Pair], k:int, random_state:int=42):
            import kmedoids
            if len(pairs_list) <= k:
                return pairs_list
            distance_matrix = np.zeros((len(pairs_list), len(pairs_list)))
            for index1, pair in enumerate(pairs_list):
                for index2 in range(index1+1, len(pairs_list)):
                    other_pair = pairs_list[index2]
                    active_text_distance = Levenshtein.distance(pair.active.text, other_pair.active.text)
                    passive_text_distance = Levenshtein.distance(pair.passive.text, other_pair.passive.text)
                    distance = active_text_distance + passive_text_distance
                    distance_matrix[index1, index2] = distance
                    distance_matrix[index2, index1] = distance

            result = kmedoids.fasterpam(distance_matrix, k, random_state=random_state, init="build")

            return [pairs_list[index] for index in result.medoids]
    
        pairs_list = self.pairs_sorted(exclude_rdgai=True)
        pairs_with_descriptions = [pair for pair in pairs_list if pair.has_description()]
        representative_pairs = []
        if pairs_with_descriptions:
            representative_pairs = find_representative_examples(pairs_with_descriptions, k, random_state=random_state)
        
        if len(representative_pairs) < k:
            pairs_without_descriptions = [pair for pair in pairs_list if not pair.has_description()]
            additional_pairs = find_representative_examples(pairs_without_descriptions, k-len(representative_pairs), random_state=random_state)
            representative_pairs.extend(additional_pairs)
        
        return representative_pairs


@dataclass
class Pair():
    active: Reading
    passive: Reading
    types: set[RelationType] = field(default_factory=set)

    def __post_init__(self):
        for relation_type in self.types:
            relation_type.pairs.add(self)

    def __str__(self):
        return f"{self.active} ➞ {self.passive}"
    
    def print(self, console):
        console.print(f"[bold red]{self.app}[/bold red]: [green]{self.active}[/green] [red]➞[/red] [green]{self.passive}[/green]")

    def reading_transition_str(self) -> str:
        return f"{self.active or 'OMISSION'} → {self.passive or 'OMISSION'}"
    
    def __repr__(self) -> str:
        return str(self)

    @property
    def app(self) -> "App":
        # assert self.active.app == self.passive.app
        return self.active.app
    
    def __hash__(self):
        return hash((self.active, self.passive))
    
    def app_element(self) -> Element:
        return find_parent(self.active.element, "app")
    
    def relation_elements(self) -> list[Element]:
        list_relation = find_element(self.app_element(), ".//listRelation[@type='transcriptional']")
        if list_relation is None:
            return []

        return find_elements(list_relation, f".//relation[@active='{self.active.n}'][@passive='{self.passive.n}']")
    
    def element_for_type(self, type:RelationType) -> Element|None:        
        for relation in self.relation_elements():
            if f"#{type.name}" in relation.attrib.get("ana").split():
                return relation
        return None
    
    def get_inverse(self) -> "Pair":
        found_pair = None
        for pair in self.app.pairs:
            if pair.active == self.passive and pair.passive == self.active:
                found_pair = pair
                break
        assert found_pair is not None, f"No inverse pair found for {self}"
        return found_pair
    
    def add_type_with_inverse(self, type:RelationType, responsible:str|None=None, description:str="", inverse_description:str="") -> Element:
        relation = self.add_type(type, responsible=responsible, description=description)
        inverse = self.get_inverse()
        inverse.add_type(type.get_inverse(), responsible=responsible, description=inverse_description)
        return relation

    def add_type(self, type:RelationType, responsible:str|None=None, description:str="") -> Element:
        self.types.add(type)
        type.pairs.add(self)

        # Check if the relation already exists
        relation = self.element_for_type(type)
        if relation is not None:
            return relation

        list_relation = find_element(self.app_element(), ".//listRelation[@type='transcriptional']")
        if list_relation is None:
            list_relation = ET.SubElement(self.app_element(), "listRelation", attrib={"type":"transcriptional"})
        
        relation = find_element(list_relation, f".//relation[@active='{self.active.n}'][@passive='{self.passive.n}']")
        if relation is not None:
            if type.name not in relation.attrib.get("ana").split():
                relation.attrib["ana"] += f" #{type.name}"
        else:
            relation = ET.SubElement(list_relation, "relation", attrib={"active":self.active.n, "passive":self.passive.n, "ana":f"#{type.name}"})

        if responsible is not None:
            relation.set("resp", responsible)

        self.add_description(description, relation)

        return relation
    
    def remove_description(self):
        for relation in self.relation_elements():
            for desc in find_elements(relation, ".//desc"):
                relation.remove(desc)
                
    def add_description(self, description:str, relation:Element|None=None):
        if relation is None:
            relation_elements = self.relation_elements()

            if len(relation_elements) == 0:
                list_relation = find_element(self.app_element(), ".//listRelation[@type='transcriptional']")
                if list_relation is None:
                    list_relation = ET.SubElement(self.app_element(), "listRelation", attrib={"type":"transcriptional"})
                relation = ET.SubElement(list_relation, "relation", attrib={"active":self.active.n, "passive":self.passive.n})
            else:        
                relation = relation_elements[0]

        description = description.strip()
        if description:
            description_element = find_element(relation, ".//desc")
            if description_element is None:
                description_element = ET.SubElement(relation, "desc")
                
            description_element.text = description

    def remove_type(self, relation_type:RelationType):
        if relation_type in self.types:
            self.types.remove(relation_type)

        if self in relation_type.pairs:    
            relation_type.pairs.remove(self)

        list_relation = find_element(self.app_element(), ".//listRelation[@type='transcriptional']")
        for relation in find_elements(list_relation, f".//relation[@active='{self.active.n}'][@passive='{self.passive.n}']"):
            if f"#{relation_type.name}" in relation.attrib.get("ana").split():
                relation.attrib['ana'] = " ".join([ana for ana in relation.attrib.get("ana").split() if ana != f"#{relation_type.name}"])
            if not relation.attrib.get("ana"):
                relation.getparent().remove(relation)

    def remove_type_with_inverse(self, relation_type:RelationType):
        self.remove_type(relation_type)
        inverse = self.get_inverse()
        inverse.remove_type(relation_type.get_inverse())

    def remove_all_types(self):
        for relation_type in set(self.types):
            self.remove_type_with_inverse(relation_type)

    def rdgai_responsible(self) -> bool:
        for element in self.relation_elements():
            if element.attrib.get('resp', '') == '#rdgai':
                return True
        return False

    def relation_type_names(self) -> set[str]:
        return set(type.name for type in self.types)
    
    def has_description(self) -> bool:
        for relation in self.relation_elements():
            if find_element(relation, ".//desc") is not None:
                return True
        return False

    def get_description(self) -> str:
        description = ""
        for relation_element in self.relation_elements():
            for desc in find_elements(relation_element, ".//desc"):
                description += "\n" + extract_text(desc)
        return description.strip()


@dataclass
class App():
    element: Element
    doc: "Doc"
    readings: list[Reading] = field(default_factory=list)
    pairs: list[Pair] = field(default_factory=list)
    non_redundant_pairs: list[Pair] = field(default_factory=list)

    def __post_init__(self):
        for reading in find_elements(self.element, ".//rdg"):
            self.readings.append(Reading(reading, app=self))

        # Build list of relation elements
        relation_elements = []
        for list_relation in find_elements(self.element, ".//listRelation[@type='transcriptional']"):
            for relation_element in find_elements(list_relation, ".//relation"):
                relation_elements.append(relation_element)
        
        # Build list of relation pairs
        active_visited = set()
        for active in self.readings:
            active_visited.add(active)
            for passive in self.readings:
                if active == passive:
                    continue

                types = set()
                for relation_element in relation_elements:
                    if relation_element.attrib.get("active") == active.n and relation_element.attrib.get("passive") == passive.n:
                        for ana in relation_element.attrib.get("ana", "").split():
                            if ana.startswith("#"):
                                ana = ana[1:]
                            if ana:
                                types.add(ana)

                pair_relation_types = set()
                for type_name in types:
                    relation_type = self.doc.relation_types[type_name] if type_name in self.doc.relation_types else self.doc.add_relation_type(type_name)                    
                    pair_relation_types.add(relation_type)

                pair = Pair(active=active, passive=passive, types=pair_relation_types)
                self.pairs.append(pair)
                if passive not in active_visited:
                    self.non_redundant_pairs.append(pair)

                for relation_type in pair_relation_types:
                    relation_type.pairs.add(pair)

        assert len(self.pairs) == len(self.non_redundant_pairs) * 2

    def get_classified_pairs(self, redundant:bool=True) -> list[Pair]:
        pairs = self.pairs if redundant else self.non_redundant_pairs
        return [pair for pair in pairs if len(pair.types) > 0]

    def get_unclassified_pairs(self, redundant:bool=True) -> list[Pair]:
        pairs = self.pairs if redundant else self.non_redundant_pairs
        return [pair for pair in pairs if len(pair.types) == 0]

    def __hash__(self):
        return hash(self.element)

    def __str__(self):
        name = self.element.attrib.get('{http://www.w3.org/XML/1998/namespace}id', '')
        if not name:
            name = self.element.attrib.get('n', '')

        if not name:
            ab = self.ab()
            for index, app in enumerate(find_elements(ab, ".//app")):
                if app == self.element:
                    name = make_nc_name(f"{self.ab_name()}-{index+1}")
                    self.element.attrib['{http://www.w3.org/XML/1998/namespace}id'] = name
                    break
        return str(name).replace(" ", "_").replace(":", "_")

    def ab(self) -> Element|None:
        return find_parent(self.element, "ab")

    def ab_name(self) -> str:
        ab = self.ab()
        return ab.attrib.get("n", "")
    
    def text_before(self) -> str:
        ab = self.ab()
        if ab is None:
            return ""
        
        items = []
        for child in ab:
            if child == self.element:
                break
            child_text = extract_text(child)
            if child_text:
                items.append(child_text)

        text = " ".join(items)
        return text.strip()
    
    def text_in_context(self, text="") -> str:
        return f"{self.text_before()} {self.text_with_signs(text)} {self.text_after()}".strip()

    def text(self) -> str:
        return extract_text(self.element)

    def text_with_signs(self, text="") -> str:
        text = text or self.text()
        if not text:
            return "⸆"
        return f"⸂{text}⸃"

    def text_after(self) -> str:
        ab = self.ab()
        if ab is None:
            return ""
        
        items = []
        reached_element = False
        for child in ab:
            if reached_element:
                child_text = extract_text(child)
                if child_text:
                    items.append(child_text)
            if child == self.element:
                reached_element = True

        text = " ".join(items)
        return text.strip()


@dataclass
class Doc():
    path: Path
    tree: ElementTree = field(default=None)
    apps: list[App] = field(default_factory=list)
    relation_types: dict[str,RelationType] = field(default_factory=dict)
    
    def __post_init__(self):
        self.tree = read_tei(self.path)
        self.relation_types = self.get_relation_types()

        for app_element in find_elements(self.tree, ".//app"):
            app = App(app_element, doc=self)
            self.apps.append(app)

    def get_interpgrp(self) -> Element:
        text = find_element(self.tree, "text") 
        interp_group = find_element(text, ".//interpGrp[@type='transcriptional']") 
        if interp_group is None: 
            interp_group = ET.Element("interpGrp", attrib={"type":"transcriptional"}) 
            text.insert(0, interp_group) 

        return interp_group

    def add_relation_type(self, name:str, description:str="") -> RelationType:
        if name in self.relation_types:
            assert self.relation_types[name].description == description, f"RelationType {name} already exists with a different description."
            return self.relation_types[name]
    
        interp_group = self.get_interpgrp()
        interp = find_element(interp_group, f".//interp[@xml:id='{name}']")
        if interp is None:
            interp = ET.Element("interp", attrib={"{http://www.w3.org/XML/1998/namespace}id":name})
            interp_group.append(interp)

        relation_type = RelationType(name=name, element=interp, description="")
        self.relation_types[name] = relation_type
        return relation_type

    def __str__(self):
        return str(self.path)

    def __repr__(self) -> str:
        return str(self)

    def write(self, output:str|Path):
        write_tei(self.tree, output)

    @property
    def language(self):
        return get_language(self.tree)
    
    def get_relation_types(self, categories_to_ignore:list[str]|None=None) -> list[RelationType]:
        interp_group = self.get_interpgrp()
        categories_to_ignore = categories_to_ignore or []
        
        relation_types = dict()
        assert interp_group is not None, "No interpGrp of type='transcriptional' found in TEI file."
        
        for interp in find_elements(interp_group, "./interp"):
            name = interp.attrib.get("{http://www.w3.org/XML/1998/namespace}id", "")
            if name in categories_to_ignore: continue

            description = extract_text(interp).strip()
            relation_types[name] = RelationType(name=name, element=interp, description=description)

        # get corresponding relations
        for category in relation_types.values():
            inverse_name = category.element.attrib.get("corresp", "")
            if inverse_name.startswith("#"):
                inverse_name = inverse_name[1:]

            if inverse_name in relation_types:
                inverse = relation_types[inverse_name]
                category.inverse = inverse
                if inverse.inverse is None:
                    inverse.inverse = category
                else:
                    assert inverse.inverse == category, f"Inverse category {inverse} already has an inverse {inverse.inverse}."

        return relation_types

    def get_classified_pairs(self, redundant:bool=True) -> list[Pair]:
        pairs = []
        for app in self.apps:
            pairs.extend(app.get_classified_pairs(redundant=redundant))

        return pairs

    def get_unclassified_pairs(self, redundant:bool=True) -> list[Pair]:
        pairs = []
        for app in self.apps:
            pairs.extend(app.get_unclassified_pairs(redundant=redundant))

        return pairs

    def print_classified_pairs(self, console:Console|None=None) -> None:
        console = console or Console()
        for relation_type in self.relation_types.values():
            console.rule(str(relation_type))
            console.print(relation_type.description, style="grey46")
            for pair in relation_type.pairs_sorted():
                pair.print(console)

            console.print("")

    def render_html(self, output:Path|None=None, all_apps:bool=False) -> str:
        from flask import Flask, request, render_template
        
        mapper = Mapper()
        app = Flask(__name__)

        with app.app_context():
            html = render_template('server.html', doc=self, mapper=mapper, all_apps=all_apps)
        
        if output:
            output.parent.mkdir(parents=True, exist_ok=True)
            output.write_text(html)
        
        return html

    def flask_app(self, output:Path, all_apps:bool=False):
        from flask import Flask, request, render_template

        self.write(output)
        mapper = Mapper()
        
        app = Flask(__name__)

        @app.route("/")
        def root():
            return render_template('server.html', doc=self, mapper=mapper, all_apps=all_apps)

        @app.route("/api/relation-type", methods=['POST'])
        def api_relation_type():
            data = request.get_json()
            
            relation_type = mapper.obj(data['relation_type'])
            assert isinstance(relation_type, RelationType), f"Expected RelationType, got {type(relation_type)}"
            
            pair = mapper.obj(data['pair'])
            assert isinstance(pair, Pair), f"Expected Pair, got {type(pair)}"

            try:
                if data['operation'] == 'remove':
                    print('remove', relation_type)
                    pair.remove_type_with_inverse(relation_type)
                elif data['operation'] == 'add':
                    print('add', relation_type)
                    pair.add_type_with_inverse(relation_type)
                else:
                    raise ValueError(f"Unknown operation {data['operation']}")
                
                print('write', output)
                self.write(output)
                return "Success", 200           
            except Exception as e:  
                print(str(e))
                return str(e), 400

            return "Failed", 400
        
        @app.route("/api/desc", methods=['POST'])
        def desc():
            data = request.get_json()
                        
            pair = mapper.obj(data['pair'])
            assert isinstance(pair, Pair), f"Expected Pair, got {type(pair)}"

            try:
                if data['operation'] == 'remove':
                    pair.remove_description()
                elif data['operation'] == 'add':
                    pair.add_description(data['description'])   
                else:
                    raise ValueError(f"Unknown operation {data['operation']}")

                print('write', output)
                self.write(output)
                return "Success", 200           
            except Exception as e:  
                print(str(e))
                return str(e), 400

            return "Failed", 400

        return app
        # app.run(debug=True, use_reloader=True)

    def clean(self, output:Path|None=None):
        """ Cleans a TEI XML file for common errors. """

        # find all listRelation elements
        list_relations = find_elements(self.tree, ".//listRelation")
        for list_relation in list_relations:
            relations_so_far = set()
            for relation in find_elements(list_relation, ".//relation"):
                # make sure that relation elements have a # at the start of the ana attribute
                if not relation.attrib['ana'].startswith("#"):
                    relation.attrib['ana'] = f"#{relation.attrib['ana']}"
                
                relations_so_far.add( (relation.attrib['active'], relation.attrib['passive']) )
            
            # consolidate duplicate relations
            for active, passive in relations_so_far:
                relations = find_elements(list_relation, f".//relation[@active='{active}'][@passive='{passive}']")
                if len(relations) > 1:
                    analytic_set = set()
                    for relation in relations:
                        analytic_set.update(relation.attrib['ana'].split())

                    for relation in relations[1:]:
                        list_relation.remove(relation)

                    relations[0].attrib['ana'] = " ".join(sorted(analytic_set))
        
        if output:
            output = Path(output)
            output.parent.mkdir(parents=True, exist_ok=True)
            print("Writing to", output)
            self.write(output)     


import numpy as np
from pathlib import Path
from dataclasses import dataclass
from langchain_core.language_models.llms import LLM
from langchain_core.output_parsers import StrOutputParser

from .tei import find_elements
from .apparatus import App, Doc, Pair
from .prompts import build_preamble, build_review_prompt

@dataclass
class EvalItem:
    app_id:str
    app:App
    active:str
    passive:str
    text_in_context:str
    reading_transition_str:str
    ground_truth:str
    predicted:str
    description:str = ""
    ground_truth_description:str = ""


def llm_review_results(
    doc:Doc,
    correct_items:list[EvalItem],
    incorrect_items:list[EvalItem],
    llm:LLM|None=None,
    examples:int=10,        
):
    template = build_review_prompt(doc, correct_items, incorrect_items, examples=examples)
    result = ""
    if llm:
        chain = template | llm | StrOutputParser()
        result = chain.invoke({})
        
    return template, result


def evaluate_docs(
    doc:Doc, 
    ground_truth:Doc,
    pairs:list[Pair]|None=None,
    confusion_matrix:Path|None=None,
    confusion_matrix_plot:Path|None=None,
    report:Path|None=None,
    llm:LLM|None=None,
    examples:int=10,
):
    # get dictionary of ground truth apps
    ground_truth_apps = {str(app):app for app in ground_truth.apps}

    # find all classified relations in the ground truth that correspond to the classified relations in the doc
    predicted = []
    gold = []
    correct_items = []
    incorrect_items = []

    # Create a dictionary of ground truth relation elements to their corresponding pairs
    ground_truth_pair_dict = dict()
    for pair in ground_truth.get_classified_pairs():
        for relation_element in pair.relation_elements():
            ground_truth_pair_dict[relation_element] = pair


    # find all classified relations in the doc that have been classified with rdgai
    pairs = pairs or [pair for pair in doc.get_classified_pairs() if pair.rdgai_responsible()]

    if len(pairs) == 0:
        print("No rdgai relations found in predicted document.")
        return

    for pair in pairs:
        # find app
        app = pair.app
        app_id = str(app)
        
        active = pair.active.n
        passive = pair.passive.n

        ground_truth_app = ground_truth_apps.get(app_id, None)
        if ground_truth_app is None:
            continue
        
        ground_truth_relations = [element for element in find_elements(ground_truth_app.element, f".//relation[@active='{active}']") if element.attrib['passive'] == passive]
        if not ground_truth_relations:
            continue
        ground_truth_relation = ground_truth_relations[0]
        ground_truth_pair = ground_truth_pair_dict[ground_truth_relation]
        
        # exclude any classified with rdgai
        if ground_truth_pair.rdgai_responsible():
            continue

        ground_truth_types = ground_truth_pair.relation_type_names()
        predicted_types = pair.relation_type_names()

        description = pair.get_description()

        eval_item = EvalItem(
            app_id=app_id,
            app=app,
            text_in_context=app.text_in_context(),
            active=ground_truth_pair.active,
            passive=ground_truth_pair.passive,
            reading_transition_str=ground_truth_pair.reading_transition_str(),
            ground_truth=ground_truth_types,
            predicted=predicted_types,
            description=description,
            ground_truth_description=ground_truth_pair.get_description(),
        )
        if ground_truth_types == predicted_types:
            correct_items.append(eval_item)
        else:
            incorrect_items.append(eval_item)

        predicted.append(" ".join(sorted(predicted_types)))
        gold.append(" ".join(sorted(ground_truth_types)))

    print(len(predicted), len(gold))
    assert len(predicted) == len(gold), f"Predicted and gold lengths do not match: {len(predicted)} != {len(gold)}"

    if len(gold) == 0:
        print("No relations found in ground truth.")
        return

    from sklearn.metrics import precision_score, recall_score, f1_score, classification_report, accuracy_score
    print(classification_report(gold, predicted))

    precision = precision_score(gold, predicted, average='macro')*100.0
    print("precision", precision)
    recall = recall_score(gold, predicted, average='macro')*100.0
    print("recall", recall)
    f1 = f1_score(gold, predicted, average='macro')*100.0
    print("f1", f1)
    accuracy = accuracy_score(gold, predicted)*100.0
    print("accuracy", accuracy)

    # create confusion matrix
    if confusion_matrix or confusion_matrix_plot or report:
        from sklearn.metrics import confusion_matrix as sk_confusion_matrix
        import pandas as pd

        labels = list(ground_truth.relation_types.keys())
        cm = sk_confusion_matrix(gold, predicted, labels=labels)
        confusion_df = pd.DataFrame(cm, index=labels, columns=labels)
        if confusion_matrix:
            confusion_matrix = Path(confusion_matrix)
            confusion_matrix.parent.mkdir(parents=True, exist_ok=True)
            confusion_df.to_csv(confusion_matrix)

        if confusion_matrix_plot or report:
            import plotly.graph_objects as go

            sums = cm.sum(axis=1, keepdims=True)
            cm_normalized = cm / np.maximum(sums, 1)

            text_annotations = [[str(cm[i][j]) for j in range(len(labels))] for i in range(len(labels))]

            # Plot the normalized confusion matrix
            fig = go.Figure(data=go.Heatmap(
                z=cm_normalized,
                x=labels,
                y=labels,
                colorscale='Viridis',
                text=text_annotations,      # Add only the raw counts to each cell
                colorbar=dict(title="Proportion of True Values")  # Updated legend title
            ))
            annotations = []
            for i in range(len(labels)):
                for j in range(len(labels)):
                    count = cm[i][j]  # Raw count
                    proportion = cm_normalized[i][j]  # Normalized proportion
                    annotations.append(
                        go.layout.Annotation(
                            x=j, y=i,
                            text=f"{count}",  # Showing both raw count and normalized proportion
                            showarrow=False,
                            font=dict(size=10, color="white" if proportion < 0.5 else "black")
                        )
                    )
            fig.update_layout(annotations=annotations)


            fig.update_layout(
                xaxis_title='Predicted',
                yaxis_title='Actual',
                xaxis=dict(tickmode='array', tickvals=list(range(len(labels))), ticktext=labels, side="top"),
                yaxis=dict(tickmode='array', tickvals=list(range(len(labels))), ticktext=labels, autorange="reversed"),
            )

            # Save the plot as HTML
            if confusion_matrix_plot:
                confusion_matrix_plot = Path(confusion_matrix_plot)
                confusion_matrix_plot.parent.mkdir(parents=True, exist_ok=True)
                fig.write_html(confusion_matrix_plot)

            if report:
                from flask import Flask, render_template
                report = Path(report)
                report.parent.mkdir(parents=True, exist_ok=True)
                app = Flask(__name__)

                review_template, review_result = llm_review_results(
                    doc, 
                    correct_items=correct_items, 
                    incorrect_items=incorrect_items, 
                    examples=examples, 
                    llm=llm,
                )

                import plotly.io as pio
                confusion_matrix_html = pio.to_html(fig, full_html=True, include_plotlyjs='inline')

                with app.app_context():
                    text = render_template(
                        'report.html', 
                        correct_items=correct_items, 
                        incorrect_items=incorrect_items, 
                        confusion_matrix=confusion_matrix_html,
                        accuracy=accuracy,
                        precision=precision,
                        recall=recall,
                        f1=f1,
                        correct_count=len(correct_items),
                        incorrect_count=len(incorrect_items),
                        prompt=build_preamble(doc, examples=examples),
                        review_template=review_template,
                        review_result=review_result,
                    )
                
                print(f"Writing HTML report to {report}")
                report.write_text(text)





import yaml

# Input and output file names
input_file = "language-subtag-registry"
output_file = "language-subtag-registry.yaml"

# Initialize variables
codes = {}

# Initialize variables
codes = {}
subtag = None
descriptions = []
current_key = None

# Read and parse the input file
with open(input_file, "r") as f:
    for line in f:
        line = line.rstrip()

        # Handle multi-line values (continuation lines start with a space)
        if line.startswith(" "):
            if current_key == "Description":
                descriptions[-1] += " " + line.strip()
            continue

        # Detect a new record
        if line == "%%":
            if subtag and descriptions:
                codes[subtag] = ", ".join(descriptions)
            subtag = None
            descriptions = []
            current_key = None
            continue

        # Extract key-value pairs
        if ": " in line:
            key, value = line.split(": ", 1)
            current_key = key.strip()

            if key == "Subtag":
                subtag = value.strip()
            elif key == "Description":
                descriptions.append(value.strip())

    # Add the last record if it exists
    if subtag and descriptions:
        codes[subtag] = ", ".join(descriptions)


# Write the YAML output
with open(output_file, "w") as f:
    f.write("# Derived from the IANA language subtag registry\n")
    f.write("# Source: https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry\n\n")
    yaml.dump(
        codes,
        f,
        sort_keys=False,  # Ensures dictionary order is maintained
        default_flow_style=False,  # Outputs YAML in a human-readable block format
    )

print(f"Converted to YAML format in {output_file}.")



from pathlib import Path
from langchain_core.output_parsers import StrOutputParser
from langchain_core.language_models.llms import LLM
import llmloader
from rich.console import Console
from rich.progress import track

from .prompts import build_template
from .parsers import CategoryParser
from .apparatus import Doc, Pair


DEFAULT_MODEL_ID = "gpt-4o"


def classify_pair(
    doc:Doc,
    pair:Pair,
    llm:LLM,
    output:Path,
    verbose:bool=False,
    prompt_only:bool=False,
    examples:int=10,
    console:Console|None=None,
):
    """
    Classifies relations for a pair of readings.
    """
    assert isinstance(doc, Doc), f"Expected Doc, got {type(doc)}"

    console = console or Console()

    template = build_template(pair, examples=examples)
    if verbose or prompt_only:
        template.pretty_print()
        if prompt_only:
            return

    chain = template | llm.bind(stop=["----"]) | StrOutputParser() | CategoryParser(doc.relation_types.keys())

    doc.write(output)

    category, description = chain.invoke({})

    console.print()
    pair.print(console)
    console.print(category, style="green bold")
    console.print(description, style="grey46")

    relation_type = doc.relation_types.get(category, None)
    if relation_type is None:
        return
    
    inverse_description = f"c.f. {pair.active} ➞ {pair.passive}"
    pair.add_type_with_inverse(relation_type, responsible="#rdgai", description=description, inverse_description=inverse_description)

    doc.write(output)
    
    
def classify(
    doc:Doc,
    output:Path,
    pairs:list[Pair]|None=None,
    verbose:bool=False,
    api_key:str="",
    llm:str=DEFAULT_MODEL_ID,
    temperature:float=0.1,
    prompt_only:bool=False,
    examples:int=10,
    console:Console|None=None,
):
    """
    Classifies relations in TEI documents.
    """
    assert isinstance(doc, Doc), f"Expected Doc, got {type(doc)}"

    console = console or Console()
    llm = llmloader.load(model=llm, api_key=api_key, temperature=temperature)
    
    pairs = pairs or doc.get_unclassified_pairs(redundant=False)
    for pair in track(pairs):
        classify_pair(doc, pair, llm, output, verbose=verbose, prompt_only=prompt_only, examples=examples, console=console)
        
        



import re
from pathlib import Path
from lxml import etree as ET
from lxml.etree import _ElementTree as ElementTree
from lxml.etree import _Element as Element

from .languages import convert_language_code


def get_language_code(doc:ElementTree|Element) -> str:
    """ Reads the element <text> and returns the value of the xml:lang attribute."""
    text = find_element(doc, ".//text")
    if text is None:
        return ""
    
    return text.attrib.get("{http://www.w3.org/XML/1998/namespace}lang", "")


def get_language(doc:ElementTree|Element) -> str:
    code = get_language_code(doc)
    return convert_language_code(code)
    

def make_nc_name(string):
    invalid_chars = "!\"#$%&'()*+/:;<=>?@[\\]^,{|}~` "
    result = string.translate(str.maketrans(invalid_chars, '_' * len(invalid_chars)))
    # if result[0].isdigit or result[0] in [".", "-"]:
    #     result = "id-" + result

    return result


def extract_text(node:Element, include_tail:bool=True) -> str:
    if node is None:
        return ""
    
    tag = re.sub(r"{.*}", "", node.tag)

    if tag in ["pc", "witDetail", "note"]:
        return ""
    if tag == "app":            
        lemma = find_element(node, ".//lem")
        if lemma is None:
            lemma = find_element(node, ".//rdg")
        return extract_text(lemma) or ""
    if tag == "ref":
        root = node.getroottree().getroot()
        target_id = node.attrib['target'].lstrip("#")
        ns = {"tei": "http://www.tei-c.org/ns/1.0"}
        target = root.xpath(f"//*[@xml:id='{target_id}']", namespaces=ns)

        if target:
            return extract_text(target[0])


    text = node.text or ""
    for child in node:
        text += " " + extract_text(child)

    if include_tail and node.tail:
        text += " " + node.tail

    return text.strip()


def read_tei(path:Path) -> ElementTree:
    parser = ET.XMLParser(remove_blank_text=True)
    with open(path, 'r') as f:
        return ET.parse(f, parser)


def find_element(doc:ElementTree|Element, xpath:str) -> Element|None:
    assert doc is not None, f"Document is None in find_element({doc}, {xpath})"
    if isinstance(doc, ElementTree):
        doc = doc.getroot()
    namespaces = doc.nsmap | {"xml": "http://www.w3.org/XML/1998/namespace"}
    element = doc.find(xpath, namespaces=namespaces)
    if element is None:
        try:
            element = doc.find(xpath)
        except SyntaxError:
            return None
    return element


def find_elements(doc:ElementTree|Element, xpath:str) -> Element|None:
    if isinstance(doc, ElementTree):
        doc = doc.getroot()
    namespaces = doc.nsmap | {"xml": "http://www.w3.org/XML/1998/namespace"}
    results = doc.findall(xpath, namespaces=namespaces)
    results += doc.findall(xpath)
    return results


def find_parent(element:Element, tag:str) -> Element|None:
    """
    Finds the nearest ancestor of the given element with the specified tag.

    Args:
        element (Element): The starting XML element from which to search upward.
        tag (str): The tag name of the ancestor element to find.

    Returns:
        Optional[Element]: The nearest ancestor element with the specified tag, or None if no such element is found.

    Example:
        >>> from xml.etree.ElementTree import Element
        >>> root = Element('root')
        >>> ab = Element('ab')
        >>> section = Element('section')
        >>> target = Element('target')
        >>> root.append(ab)
        >>> ab.append(section)
        >>> section.append(target)
        >>> result = find_parent(target, 'ab')
        >>> assert result == ab

        This will find the <ab> ancestor of the <target> element.
    """
    while element is not None:
        element_tag = re.sub(r"{.*}", "", element.tag)
        if element_tag == tag:
            return element
        element = element.getparent()
    return None


def write_tei(doc:ElementTree, path:Path|str) -> None:
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    doc.write(str(path), encoding="utf-8", xml_declaration=True, pretty_print=True)


def get_reading_identifier(reading:Element, check:bool=False, create_if_necessary:bool=True) -> str:
    identifier = reading.attrib.get("{http://www.w3.org/XML/1998/namespace}id", "")
    if not identifier:
        identifier = reading.attrib.get("n", "")

    if not identifier and create_if_necessary:
        app = reading.getparent()
        identifier = 1
        while find_element(app, f".//rdg[@n='{identifier}']") is not None:
            identifier += 1
        identifier = str(identifier)
        reading.attrib["n"] = identifier
    
    if check:
        assert identifier, f"Reading {reading} must have a name attribute 'xml:id' or 'n'."
    
    return identifier
